/*
    This file is part of magpar.

    Copyright (C) 2002-2010 Werner Scholz

    www:   http://www.magpar.net/
    email: magpar(at)magpar.net

    magpar is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    magpar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with magpar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static char const Id[] = "$Id: mesh2dual.c 2962 2010-02-04 19:50:44Z scholz $\n\n";
static char const Td[] = "$Today: " __FILE__ "  " __DATE__ "  "  __TIME__  " $\n\n";

#include "util.h"

#ifdef METIS
EXTERN_C_BEGIN
#include "metis.h"
EXTERN_C_END

#ifdef DEBUG
/* for size of connectivity matrix generated by METIS_MeshToNodal */
#define JASIZE 25
#endif
#endif

int Mesh2VertEle(int n_ele, int n_vert, int *elevert, int **ia, int **ja)
{
  MagparFunctionLogBegin;

  /* This function calculates the vertex-element connectivity.
   * For each vertex all elements, which share this vertex are stored.
   *
   * This is different from METIS's MeshToNodal, which calculates
   * the connectivity between nodes (i.e. it connects two vertices [nodes] in
   * the graph if they are connected by an edge in the mesh).
   */

  int *tca;
  ierr = PetscMalloc(n_vert*sizeof(int),&tca);CHKERRQ(ierr);
  ierr = PetscMemzero(tca,n_vert*sizeof(int));CHKERRQ(ierr);

  /* count number of elements sharing each node */
  for (int i=0;i<n_ele;i++) {
    for (int j=0;j<NV;j++) {
      tca[elevert[NV*i+j]]++;
    }
  }

  /* create index array */
  int *tia;
  ierr = PetscMalloc((n_vert+1)*sizeof(int),&tia);CHKERRQ(ierr);
  tia[0]=0;
  for (int i=1;i<n_vert+1;i++) {
    tia[i]=tia[i-1]+tca[i-1];
  }

  PetscPrintf(PETSC_COMM_WORLD,
    "Allocating %i MB for %i elements in adjacency matrix\n",
    tia[n_vert]*sizeof(int)/1024/1024+1,
    tia[n_vert]
  );

  int *tja;
  ierr = PetscMalloc(tia[n_vert]*sizeof(int),&tja);CHKERRQ(ierr);

  ierr = PetscMemzero(tca,n_vert*sizeof(int));CHKERRQ(ierr);

  /* find all elements, which share a node */
  for (int i=0;i<n_ele;i++) {
    for (int j=0;j<NV;j++) {
      int v;
      v=elevert[NV*i+j];
      tja[tia[v]+tca[v]]=i;
      tca[v]++;
    }
  }

  /* count total number of connectivities */
  int cnt=0;
  for (int i=0;i<n_vert;i++) {
    cnt+=tca[i];
    assert(tca[i]==tia[i+1]-tia[i]);
  }
  assert(cnt==tia[n_vert]);

#ifdef DEBUG
  PetscPrintf(PETSC_COMM_WORLD,"Dual mesh:\n");
  PetscPrintf(PETSC_COMM_WORLD,"   node  |   #  | elements; ia\n");

  int cnt2=0;
  for (int i=0;i<n_vert;i++) {
    PetscPrintf(PETSC_COMM_WORLD,"%7i %6i ",i,tia[i+1]-tia[i]);
    for (int j=tia[i];j<tia[i+1];j++) {
      PetscPrintf(PETSC_COMM_WORLD,"%7i ",tja[j]);
      cnt2++;
    }
    PetscPrintf(PETSC_COMM_WORLD,"%6i\n",tia[i]);
  }
  PetscPrintf(PETSC_COMM_WORLD,"%7i %6i %6i\n",n_vert,0,tia[n_vert]);

  assert(cnt==cnt2);
#endif

  ierr = PetscFree(tca);CHKERRQ(ierr);

  *ia=tia;
  *ja=tja;

  MagparFunctionLogReturn(0);
}


int Mesh2Nodal(int n_ele, int n_vert, int *elevert, int **ia, int **ja)
{
  MagparFunctionLogBegin;

  /* This function calculates the vertex-vertex connectivity.
   * For each vertex all vertices, which are connected by an edge, are stored.
   *
   * This is (for magpar's purposes) a full replacement for METIS's MeshToNodal function.
   */

  /* obtain vertex-element connectivity first */
  int *veia,*veja;
  ierr = Mesh2VertEle(n_ele,n_vert,elevert,&veia,&veja);CHKERRQ(ierr);

  /* allocate memory for counters */
  int *tca;
  ierr = PetscMalloc(n_vert*sizeof(int),&tca);CHKERRQ(ierr);
  ierr = PetscMemzero(tca,n_vert*sizeof(int));CHKERRQ(ierr);

  /* allocate temp. memory for neighbors */
#define MAXVERTVERT 10000
  int *tj;
  ierr = PetscMalloc(MAXVERTVERT*sizeof(int),&tj);CHKERRQ(ierr);

  /* count number of neighboring vertices */
  for (int v=0;v<n_vert;v++) {
    for (int i=0;i<MAXVERTVERT;i++) {
      tj[i]=-1;
    }
    for (int i=veia[v];i<veia[v+1];i++) {
      int el;
      el=veja[i];
      for (int j=0;j<NV;j++) {
        int vel;
        vel=elevert[NV*el+j];

        /* skip self */
        if (vel==v) continue;

        int found;
        found=0;
        for (int k=0;k<tca[v];k++) {
          if (vel==tj[k]) {
            found=1;
            break;
          }
        }
        if (!found) {
          /* add new vertex */
          tj[tca[v]]=vel;
          tca[v]++;
          /* check for array overrun */
          assert(tca[v]<MAXVERTVERT);
        }
      }
    }
  }

  ierr = PetscFree(tj);CHKERRQ(ierr);

  /* create index array */
  int *tia;
  ierr = PetscMalloc((n_vert+1)*sizeof(int),&tia);CHKERRQ(ierr);
  tia[0]=0;
  for (int i=1;i<n_vert+1;i++) {
    tia[i]=tia[i-1]+tca[i-1];
  }

  /* reset counter array */
  ierr = PetscMemzero(tca,n_vert*sizeof(int));CHKERRQ(ierr);

  PetscPrintf(PETSC_COMM_WORLD,
    "Allocating %i MB for %i elements in adjacency matrix\n",
    tia[n_vert]*sizeof(int)/1024/1024+1,
    tia[n_vert]
  );

  int *tja;
  ierr = PetscMalloc(tia[n_vert]*sizeof(int),&tja);CHKERRQ(ierr);
  for (int i=0;i<tia[n_vert];i++) {
    tja[i]=-1;
  }

  /* find all neighboring vertices */
  for (int v=0;v<n_vert;v++) {
    for (int i=veia[v];i<veia[v+1];i++) {
      int el;
      el=veja[i];
      for (int j=0;j<NV;j++) {
        int vel;
        vel=elevert[NV*el+j];

        /* skip self */
        if (vel==v) continue;

        int found;
        found=0;
        for (int k=tia[v];k<tia[v+1];k++) {
          if (vel==tja[k]) {
            found=1;
            break;
          }
        }
        if (!found) {
          /* add new vertex */
          tja[tia[v]+tca[v]]=vel;
          tca[v]++;
          /* check for array overrun */
          assert(tca[v]<=tia[v+1]-tia[v]);
        }
      }
    }
    assert(tca[v]==tia[v+1]-tia[v]);
  }

  /* count total number of connectivities */
  int cnt=0;
  for (int i=0;i<n_vert;i++) {
    cnt+=tca[i];
    assert(tca[i]==tia[i+1]-tia[i]);
  }
  assert(cnt==tia[n_vert]);

#ifdef DEBUG
  PetscPrintf(PETSC_COMM_WORLD,"Nodal mesh:\n");
  PetscPrintf(PETSC_COMM_WORLD,"   node  |   #  | nodes; ia\n");

  int cnt2=0;
  for (int i=0;i<n_vert;i++) {
    PetscPrintf(PETSC_COMM_WORLD,"%7i %6i ",i,tia[i+1]-tia[i]);
    for (int j=tia[i];j<tia[i+1];j++) {
      PetscPrintf(PETSC_COMM_WORLD,"%7i ",tja[j]);
      cnt2++;
    }
    PetscPrintf(PETSC_COMM_WORLD,"%6i\n",tia[i]);
  }
  PetscPrintf(PETSC_COMM_WORLD,"%7i %6i %6i\n",n_vert,0,tia[n_vert]);

  assert(cnt==cnt2);
#endif

  ierr = PetscFree(tca);CHKERRQ(ierr);
  ierr = PetscFree(veia);CHKERRQ(ierr);
  ierr = PetscFree(veja);CHKERRQ(ierr);

  *ia=tia;
  *ja=tja;

#ifdef METIS
#ifdef DEBUG
  /* compare with Metis's METIS_MeshToNodal function */
  idxtype *mia,*mja;
  ierr = PetscMalloc((n_vert+1)*sizeof(idxtype),&mia);CHKERRQ(ierr);
  ierr = PetscMalloc(JASIZE*n_vert*sizeof(idxtype),&mja);CHKERRQ(ierr);
  int  zero=0,two=2;
  METIS_MeshToNodal(
    &n_ele,
    &n_vert,
    elevert,
    &two,
    &zero,
    mia,
    mja
  );
  /* check for array overrun */
  assert(mia[n_vert]<JASIZE*n_vert);

  if (tia[n_vert]!=mia[n_vert]) {
    PetscPrintf(PETSC_COMM_WORLD,"ERROR1: %i %i %i %i %i\n",n_vert,tia[n_vert],mia[n_vert]);
    SETERRQ(PETSC_ERR_LIB,"Inconsistency with METIS detected! Exiting!\n");
  }
  for (int v=0;v<n_vert+1;v++) {
    if (tia[v]!=mia[v]) {
      PetscPrintf(PETSC_COMM_WORLD,"ERROR2: %i %i %i\n",v,tia[v],mia[v]);
      SETERRQ(PETSC_ERR_LIB,"Inconsistency with METIS detected! Exiting!\n");
    }
    for (int i=tia[v];i<tia[n_vert];i++) {
      if (tja[i]!=mja[i]) {
        PetscPrintf(PETSC_COMM_WORLD,"ERROR3: %i %i %i %i %i\n",v,tia[v],mia[v],tja[i],mja[i]);
        SETERRQ(PETSC_ERR_LIB,"Inconsistency with METIS detected! Exiting!\n");
      }
    }
  }

  ierr = PetscFree(mia);CHKERRQ(ierr);
  ierr = PetscFree(mja);CHKERRQ(ierr);
#endif
#endif

  MagparFunctionLogReturn(0);
}


int Mesh2Dual(int n_ele, int n_vert, int *elevert, int **ia, int **ja)
{
  MagparFunctionLogBegin;

  /* This function calculates the element-element connectivity.
   * For each element all elements, which share a face, are stored.
   *
   * This is (for magpar's purposes) a full replacement for METIS's MeshToDual function.
   */

  /* obtain vertex-element connectivity first */
  int *veia,*veja;
  ierr = Mesh2VertEle(n_ele,n_vert,elevert,&veia,&veja);CHKERRQ(ierr);

  /* allocate memory for counters */
  int *tca;
  ierr = PetscMalloc(n_ele*sizeof(int),&tca);CHKERRQ(ierr);
  ierr = PetscMemzero(tca,n_ele*sizeof(int));CHKERRQ(ierr);

  /* allocate temp. memory for neighbors */
#define MAXELEELE NF
  int *tj;
  ierr = PetscMalloc(MAXELEELE*sizeof(int),&tj);CHKERRQ(ierr);

  /* count number of neighboring elements */
  for (int el=0;el<n_ele;el++) {
    for (int i=0;i<MAXELEELE;i++) {
      tj[i]=-1;
    }
    for (int j=0;j<NV;j++) {
      if (tca[el]==NF) break;
      int v;
      v=elevert[NV*el+j];
      for (int i=veia[v];i<veia[v+1];i++) {
        if (tca[el]==NF) break;
        if (j==0 && tca[el]==NF-1) break;
        int nel;
        nel=veja[i];
        if (nel==el) continue;
        int x;
        x=tetnb(elevert+NV*el,elevert+NV*nel,PETSC_NULL);
        assert(x<NV);
        if (x<3) continue;

        int found;
        found=0;

        for (int k=0;k<tca[el];k++) {
          if (nel==tj[k]) {
            found=1;
            break;
          }
        }
        if (!found) {
          /* add new neighbor */
          tj[tca[el]]=nel;
          tca[el]++;
          /* check for array overrun */
          assert(tca[el]<=MAXELEELE);
          if (tca[el]==NF) break;
        }
      }
    }
  }

  ierr = PetscFree(tj);CHKERRQ(ierr);

  /* create index array */
  int *tia;
  ierr = PetscMalloc((n_ele+1)*sizeof(int),&tia);CHKERRQ(ierr);
  tia[0]=0;
  for (int i=1;i<n_ele+1;i++) {
    tia[i]=tia[i-1]+tca[i-1];
  }

  /* reset counter array */
  ierr = PetscMemzero(tca,n_ele*sizeof(int));CHKERRQ(ierr);

  PetscPrintf(PETSC_COMM_WORLD,
    "Allocating %i MB for %i elements in adjacency matrix\n",
    tia[n_ele]*sizeof(int)/1024/1024+1,
    tia[n_ele]
  );

  int *tja;
  ierr = PetscMalloc(tia[n_ele]*sizeof(int),&tja);CHKERRQ(ierr);
  for (int i=0;i<tia[n_ele];i++) {
    tja[i]=-1;
  }

  /* store all neighboring elements */
  for (int el=0;el<n_ele;el++) {
    for (int j=0;j<NV;j++) {
      if (tca[el]==NF) break;
      int v;
      v=elevert[NV*el+j];
      for (int i=veia[v];i<veia[v+1];i++) {
        if (j==0 && tca[el]==NF-1) break;
        if (tca[el]==NF) break;
        int nel;
        nel=veja[i];
        if (nel==el) continue;
        int x;
        x=tetnb(elevert+NV*el,elevert+NV*nel,PETSC_NULL);
        assert(x<NV);
        if (x<3) continue;

        int found;
        found=0;

        for (int k=0;k<tca[el];k++) {
          if (nel==tja[tia[el]+k]) {
            found=1;
            break;
          }
        }
        if (!found) {
          /* add new neighbor */
          tja[tia[el]+tca[el]]=nel;
          tca[el]++;
          /* check for array overrun */
          assert(tca[el]<=MAXELEELE);
          if (tca[el]==NF) break;
        }
      }
    }
    assert(tca[el]==tia[el+1]-tia[el]);
  }

  /* count total number of connectivities */
  int cnt=0;
  for (int i=0;i<n_ele;i++) {
    cnt+=tca[i];
    assert(tca[i]==tia[i+1]-tia[i]);
  }
  assert(cnt==tia[n_ele]);

#ifdef DEBUG
  PetscPrintf(PETSC_COMM_WORLD,"Element connectivity:\n");
  PetscPrintf(PETSC_COMM_WORLD,"   element  |   #  | neighbors; ia\n");

  int cnt2=0;
  for (int i=0;i<n_ele;i++) {
    PetscPrintf(PETSC_COMM_WORLD,"%7i %6i ",i,tia[i+1]-tia[i]);
    for (int j=tia[i];j<tia[i+1];j++) {
      PetscPrintf(PETSC_COMM_WORLD,"%7i ",tja[j]);
      cnt2++;
    }
    PetscPrintf(PETSC_COMM_WORLD,"%6i\n",tia[i]);
  }
  PetscPrintf(PETSC_COMM_WORLD,"%7i %6i %6i\n",n_ele,0,tia[n_ele]);

  assert(cnt==cnt2);
#endif

  ierr = PetscFree(tca);CHKERRQ(ierr);
  ierr = PetscFree(veia);CHKERRQ(ierr);
  ierr = PetscFree(veja);CHKERRQ(ierr);

  *ia=tia;
  *ja=tja;

#ifdef METIS
#ifdef DEBUG
  /* adjacency matrices */
  idxtype *mia,*mja;
  ierr = PetscMalloc((n_ele+1)*sizeof(idxtype),&mia);CHKERRQ(ierr);
  ierr = PetscMalloc(JASIZE*n_ele*sizeof(idxtype),&mja);CHKERRQ(ierr);
  int  zero=0,two=2;
  METIS_MeshToDual(
    &n_ele,
    &n_vert,
    elevert,
    &two,
    &zero,
    mia,
    mja
  );
  /* check for array overrun */
  assert(mia[n_ele]<NV*n_ele);

  if (tia[n_ele]!=mia[n_ele]) {
    PetscPrintf(PETSC_COMM_WORLD,"ERROR1: %i %i %i %i %i\n",n_ele,tia[n_ele],mia[n_ele]);
    SETERRQ(PETSC_ERR_LIB,"Inconsitency with METIS detected! Exiting!\n");
  }
  for (int v=0;v<n_ele+1;v++) {
    if (tia[v]!=mia[v]) {
      PetscPrintf(PETSC_COMM_WORLD,"ERROR2: %i %i %i\n",v,tia[v],mia[v]);
      SETERRQ(PETSC_ERR_LIB,"Inconsitency with METIS detected! Exiting!\n");
    }
/* METIS uses different ordering!
    for (int i=tia[v];i<tia[n_ele];i++) {
      if (tja[i]!=mja[i]) {
        PetscPrintf(PETSC_COMM_WORLD,"ERROR3: %i %i %i %i %i\n",v,tia[v],mia[v],tja[i],mja[i]);
        SETERRQ(PETSC_ERR_LIB,"Inconsitency with METIS detected! Exiting!\n");
      }
    }
*/
  }

  ierr = PetscFree(mia);CHKERRQ(ierr);
  ierr = PetscFree(mja);CHKERRQ(ierr);
#endif
#endif


  MagparFunctionLogReturn(0);
}



int Mesh2Dual2(int n_ele, int n_vert, int *elevert, int **ia, int **ja)
{
  MagparFunctionLogBegin;

  /* This function calculates the element-element connectivity.
   * For each element all elements, which share a node (!), are stored.
   *
   * This is different from Mesh2Dual!!!
   */

  /* obtain vertex-element connectivity first */
  int *veia,*veja;
  ierr = Mesh2VertEle(n_ele,n_vert,elevert,&veia,&veja);CHKERRQ(ierr);

  /* allocate memory for counters */
  int *tca;
  ierr = PetscMalloc(n_ele*sizeof(int),&tca);CHKERRQ(ierr);
  ierr = PetscMemzero(tca,n_ele*sizeof(int));CHKERRQ(ierr);

  /* allocate temp. memory for neighbors */
#define MAXELEELE NF
  int *tj;
  ierr = PetscMalloc(MAXELEELE*sizeof(int),&tj);CHKERRQ(ierr);

  /* count number of neighboring elements */
  for (int el=0;el<n_ele;el++) {
    for (int i=0;i<MAXELEELE;i++) {
      tj[i]=-1;
    }
    for (int j=0;j<NV;j++) {
      int v;
      v=elevert[NV*el+j];
      for (int i=veia[v];i<veia[v+1];i++) {
        int nel;
        nel=veja[i];
        if (nel==el) continue;
#if 0
        /* not necessary since all neighbors will be counted */
        int x;
        x=tetnb(elevert+NV*el,elevert+NV*nel,PETSC_NULL);
        assert(x<NV);
        if (x<1) continue;
#endif

        int found;
        found=0;

        for (int k=0;k<tca[el];k++) {
          if (nel==tj[k]) {
            found=1;
            break;
          }
        }
        if (!found) {
          /* add new neighbor */
          tj[tca[el]]=nel;
          tca[el]++;
          /* check for array overrun */
          assert(tca[el]<=MAXELEELE);
        }
      }
    }
  }

  ierr = PetscFree(tj);CHKERRQ(ierr);

  /* create index array */
  int *tia;
  ierr = PetscMalloc((n_ele+1)*sizeof(int),&tia);CHKERRQ(ierr);
  tia[0]=0;
  for (int i=1;i<n_ele+1;i++) {
    tia[i]=tia[i-1]+tca[i-1];
  }

  /* reset counter array */
  ierr = PetscMemzero(tca,n_ele*sizeof(int));CHKERRQ(ierr);

  PetscPrintf(PETSC_COMM_WORLD,
    "Allocating %i MB for %i elements in adjacency matrix\n",
    tia[n_ele]*sizeof(int)/1024/1024+1,
    tia[n_ele]
  );

  int *tja;
  ierr = PetscMalloc(tia[n_ele]*sizeof(int),&tja);CHKERRQ(ierr);
  for (int i=0;i<tia[n_ele];i++) {
    tja[i]=-1;
  }

  /* store all neighboring elements */
  for (int el=0;el<n_ele;el++) {
    for (int j=0;j<NV;j++) {
      int v;
      v=elevert[NV*el+j];
      for (int i=veia[v];i<veia[v+1];i++) {
        int nel;
        nel=veja[i];
        if (nel==el) continue;
#if 0
        /* not necessary since all neighbors will be counted */
        int x;
        x=tetnb(elevert+NV*el,elevert+NV*nel,PETSC_NULL);
        assert(x<NV);
        if (x<1) continue;
#endif

        int found;
        found=0;

        for (int k=0;k<tca[el];k++) {
          if (nel==tja[tia[el]+k]) {
            found=1;
            break;
          }
        }
        if (!found) {
          /* add new neighbor */
          tja[tia[el]+tca[el]]=nel;
          tca[el]++;
          /* check for array overrun */
          assert(tca[el]<=MAXELEELE);
        }
      }
    }
    assert(tca[el]==tia[el+1]-tia[el]);
  }

  /* count total number of connectivities */
  int cnt=0;
  for (int i=0;i<n_ele;i++) {
    cnt+=tca[i];
    assert(tca[i]==tia[i+1]-tia[i]);
  }
  assert(cnt==tia[n_ele]);

#ifdef DEBUG
  PetscPrintf(PETSC_COMM_WORLD,"Element connectivity:\n");
  PetscPrintf(PETSC_COMM_WORLD,"   element  |   #  | neighbors; ia\n");

  int cnt2=0;
  for (int i=0;i<n_ele;i++) {
    PetscPrintf(PETSC_COMM_WORLD,"%7i %6i ",i,tia[i+1]-tia[i]);
    for (int j=tia[i];j<tia[i+1];j++) {
      PetscPrintf(PETSC_COMM_WORLD,"%7i ",tja[j]);
      cnt2++;
    }
    PetscPrintf(PETSC_COMM_WORLD,"%6i\n",tia[i]);
  }
  PetscPrintf(PETSC_COMM_WORLD,"%7i %6i %6i\n",n_ele,0,tia[n_ele]);

  assert(cnt==cnt2);
#endif

  ierr = PetscFree(tca);CHKERRQ(ierr);
  ierr = PetscFree(veia);CHKERRQ(ierr);
  ierr = PetscFree(veja);CHKERRQ(ierr);

  *ia=tia;
  *ja=tja;

#ifdef METIS
#ifdef DEBUG
  /* adjacency matrices */
  idxtype *mia,*mja;
  ierr = PetscMalloc((n_ele+1)*sizeof(idxtype),&mia);CHKERRQ(ierr);
  ierr = PetscMalloc(JASIZE*n_ele*sizeof(idxtype),&mja);CHKERRQ(ierr);
  int  zero=0,two=2;
  METIS_MeshToDual(
    &n_ele,
    &n_vert,
    elevert,
    &two,
    &zero,
    mia,
    mja
  );
  /* check for array overrun */
  assert(mia[n_ele]<NV*n_ele);

  if (tia[n_ele]!=mia[n_ele]) {
    PetscPrintf(PETSC_COMM_WORLD,"ERROR1: %i %i %i %i %i\n",n_ele,tia[n_ele],mia[n_ele]);
    SETERRQ(PETSC_ERR_LIB,"Inconsitency with METIS detected! Exiting!\n");
  }
  for (int v=0;v<n_ele+1;v++) {
    if (tia[v]!=mia[v]) {
      PetscPrintf(PETSC_COMM_WORLD,"ERROR2: %i %i %i\n",v,tia[v],mia[v]);
      SETERRQ(PETSC_ERR_LIB,"Inconsitency with METIS detected! Exiting!\n");
    }
/* METIS uses different ordering!
    for (int i=tia[v];i<tia[n_ele];i++) {
      if (tja[i]!=mja[i]) {
        PetscPrintf(PETSC_COMM_WORLD,"ERROR3: %i %i %i %i %i\n",v,tia[v],mia[v],tja[i],mja[i]);
        SETERRQ(PETSC_ERR_LIB,"Inconsitency with METIS detected! Exiting!\n");
      }
    }
*/
  }

  ierr = PetscFree(mia);CHKERRQ(ierr);
  ierr = PetscFree(mja);CHKERRQ(ierr);
#endif
#endif


  MagparFunctionLogReturn(0);
}

